# Server -> Bridge
CALL_MESSAGE_TYPE = "FUNCTION_CALL"
GET_HOW_TO_SERVICE_REGISTRATION = "GET_HOW_TO_SERVICE_REGISTRATION"
REGISTRATION_MESSAGE = "REGISTRATION"
OAUTH_RETURN = "OAUTH_RETURN"
DATA_CALLBACK = "CALLBACK"
ADVICE_RESPONSE = "ADVICE_RESPONSE"
ADVICE_NOTIFICATION = "ADVICE_NOTIFICATION"
ICON_UPLOAD = "ICON_UPLOAD"
ESTABLISH_CONNECTION = "ESTABLISH_CONNECTION"

# Bridge -> Server
NOTIFICATION = "NOTIFICATION"
CONFIGURATION = "CONFIGURATION"
ADVICE = "ADVICE_SET"
ICON_REQUEST = "ICON_REQUEST"
